# Commands

# License {{{1
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .conceal import Conceal
from .config import read_config, get_setting, override_setting
from .editors import GenericEditor
from .generator import PasswordGenerator
from .gpg import GnuPG
from .utilities import two_columns
from .writer import get_writer
from .__init__ import __version__
from inform import Error, error, codicil, output, conjoin, os_error
from shlib import to_path, mv
from docopt import docopt
from textwrap import dedent
import re
import sys


# Utilities {{{1
def title(text):
    return text[0].upper() + text[1:] + '.'

# Command base class {{{1
class Command:
    @classmethod
    def commands(cls):
        for cmd in cls.__subclasses__():
            yield cmd
            for sub in cmd.commands():
                yield sub

    @classmethod
    def commands_sorted(cls):
        for cmd in sorted(cls.commands(), key=lambda c: c.get_name()):
            yield cmd

    @classmethod
    def find(cls, name):
        for command in cls.commands():
            if name in command.NAMES:
                return command

    @classmethod
    def execute(cls, name, args):
        command = cls.find(name)
        if command:
            command.run(name, args if args else [])
        else:
            error('unknown command.', culprit=name)
            codicil("Use 'avendesora help' for list of available commands."),

    @classmethod
    def summarize(cls, width=16):
        summaries = []
        for cmd in Command.commands_sorted():
            summaries.append(two_columns(', '.join(cmd.NAMES), cmd.DESCRIPTION))
        return '\n'.join(summaries)

    @classmethod
    def get_name(cls):
        return cls.NAMES[0]


# Add {{{1
class Add(Command):
    NAMES = 'add',
    DESCRIPTION = 'add a new account'
    USAGE = dedent("""
        Usage:
            avendesora [options] add [<template>]

        Options:
            -f <file>, --file <file>
                                    Add account to specified file.
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}

            The default template is {default}. The available templates are:
            {templates}
        """).strip()
        return text.format(
            title=title(cls.DESCRIPTION), usage=cls.USAGE,
            default=get_setting('default_account_template'),
            templates=conjoin(sorted(get_setting('account_templates').keys()))
        )

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        try:
            # get the specified template
            templates = get_setting('account_templates')
            if cmdline['<template>']:
                template_name = cmdline['<template>']
            else:
                template_name = get_setting('default_account_template')
            template = dedent(templates[template_name]).strip() + '\n'

            # save template to tmp file and open it in the editor
            from tempfile import mktemp
            tmpfile = GnuPG(mktemp(suffix='_avendesora.gpg'))
            tmpfile.save(template, get_setting('gpg_ids'))
            GenericEditor.open_and_search(tmpfile.path)

            # read the tmp file and determine if it has changed
            new = tmpfile.read()
            tmpfile.remove()
            if new == template:
                return output('Unchanged, and so ignored.')

            # hide the files that should be hidden
            def hide(match):
                return 'Hidden(%r)' % Conceal.hide(match.group(1))
            new = re.sub("<<(.*?)>>", hide, new)

            # determine the accounts file
            prefix = cmdline['--file']
            if prefix:
                candidates = [
                    p
                    for p in get_setting('accounts_files')
                    if p.startswith(prefix)
                ]
                if not candidates:
                    raise Error('not found.', cuplrit=cmdline['--file'])
                if len(candidates) > 0:
                    raise Error(
                        'ambiguous, matches %s.' % conjoin(candidates),
                        cuplrit=prefix
                    )
                filename = candidates[0]
            else:
                filename = get_setting('accounts_files')[0]
            path = to_path(get_setting('settings_dir'), filename)

            # get original contents of accounts file
            orig_accounts_file = GnuPG(path)
            accounts = orig_accounts_file.read()

            # add new account to the contents
            accounts = accounts + new + '\n'

            # rename the original file and then save the new version
            orig_accounts_file.rename('.saved')
            new_accounts_file = GnuPG(path)
            new_accounts_file.save(accounts)
        except OSError as err:
            error(os_error(err))
        except KeyError as err:
            error(
                'unknown account template, choose from %s.' % conjoin(
                    sorted(templates.keys())
                ), culprit=template_name
            )


# Browse {{{1
class Browse(Command):
    NAMES = 'browse',
    DESCRIPTION = 'open account URL in web browser'
    USAGE = dedent("""
        Usage:
            avendesora [options] browse <account> [<key>]

        Options:
            -b <browser>, --browser <browser>
                                    Open account in specified browser.
    """).strip()

    @classmethod
    def help(cls):
        default_browser = get_setting('default_browser')
        browsers = get_setting('browsers')
        browsers = '\n'.join([
            two_columns(k, browsers[k].split()[0], width=1)
            for k in sorted(browsers)
        ])
        text = dedent("""
            {title}

            {usage}

            The default browser is {default}. The available browsers are:
            {browsers}
        """).strip()
        return text.format(
            title=title(cls.DESCRIPTION), usage=cls.USAGE,
            default=default_browser, browsers=browsers
        )

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator()

        # determine the account and open the URL
        account = generator.get_account(cmdline['<account>'])
        account.open_browser(cmdline['--browser'])

# Edit {{{1
class Edit(Command):
    NAMES = 'edit',
    DESCRIPTION = 'edit an account'
    USAGE = dedent("""
        Usage:
            avendesora edit <account>
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator()

        # determine the account and open the URL
        account = generator.get_account(cmdline['<account>'])
        filepath = account._file_info.path
        account_name = account.__name__
        GenericEditor.open_and_search(filepath, account_name)


# Find {{{1
class Find(Command):
    NAMES = 'find',
    DESCRIPTION = 'find an account'
    USAGE = dedent("""
        Find accounts whose name contains the search text.

        Usage:
            avendesora find <text>
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator()

        # find accounts whose name matches the criteria
        to_print = []
        for acct in generator.find_accounts(cmdline['<text>']):
            aliases = getattr(acct, 'aliases', [])

            aliases = ' (%s)' % (', '.join(aliases)) if aliases else ''
            to_print += [acct.get_name() + aliases]
        output(cmdline['<text>']+ ':')
        output('    ' + ('\n    '.join(sorted(to_print))))


# Help {{{1
class Help(Command):
    NAMES = 'help',
    DESCRIPTION = 'give information about commands or other topics'
    USAGE = dedent("""
        Usage:
            avendesora help [<topic>]
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        from .help import HelpMessage
        HelpMessage.show(cmdline['<topic>'])


# Hide {{{1
class Hide(Command):
    NAMES = 'hide', 'conceal',
    DESCRIPTION = 'conceal text by encoding it'
    USAGE = dedent("""
        Usage:
            avendesora [options] hide <text>
            avendesora [options] conceal <text>

        Options:
            -e <encoding>, --encoding <encoding>
                                    Encoding used when concealing information.
    """).strip()

    @classmethod
    def help(cls):
        encodings = '    ' + '\n    '.join(Conceal.encodings())
        text = dedent("""
            {title}

            {usage}

            Possible encodings include:
            {encodings}

            base64 (default):
                This encoding obscures but does not encrypt the text. It can
                protect text from observers that get a quick glance of the
                encoded text, but if they are able to capture it they can easily
                decode it.
        """).strip()
        return text.format(
            title=title(cls.DESCRIPTION), usage=cls.USAGE, encodings=encodings
        )

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # transform and output the string
        output(Conceal.hide(cmdline['<text>'], cmdline['--encoding']))


# Initialize {{{1
class Initialize(Command):
    NAMES = 'init', 'initialize',
    DESCRIPTION = 'create initial set of Avendesora files'
    USAGE = dedent("""
        Usage:
            avendesora init [--gpg-id <id>]... [options]
            avendesora initialize [--gpg-id <id>]... [options]

        Options:
            -g <id>, --gpg-id <id>  Use this ID when creating any missing encrypted files.
            -h <path>, --gpg-home <path>
                                    GPG home directory (default is ~/.gnupg).
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}

            Initial configuration and accounts files are created only if they
            do not already exist.  Existing files are not modified.
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)
        if cmdline['--gpg-home']:
            override_setting('gpg_home', cmdline['--gpg-home'])

        # save the gpg_ids for the logfile in case it is encrypted.
        gpg_ids = cmdline['--gpg-id']
        if not get_setting('gpg_ids'):
            override_setting('gpg_ids', gpg_ids)

        # run the generator
        generator = PasswordGenerator(init=True, gpg_ids=gpg_ids)


# New {{{1
class New(Command):
    NAMES = 'new',
    DESCRIPTION = 'create new accounts file'
    USAGE = dedent("""
        Usage:
            avendesora new [--gpg-id <id>]... <name>

        Options:
            -g <id>, --gpg-id <id>  Use this ID when creating any missing encrypted files.
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}

            Creates a new accounts file. Accounts that share the same file share
            the same master password by default and, if the file is encrypted,
            can be decrypted by the same recipients.

            Generally you would create a new accounts file for each person or
            group with which you wish to share accounts. You would use separate
            files for passwords with different security domains. For example, a
            high-value passwords might be placed in an encrypted file that would
            only be placed highly on protected computers. Conversely, low-value
            passwords might be contained in perhaps an unencrypted file that is
            found on many computers.

            Add a '.gpg' extension to <name> to encrypt the file.
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # get the gpg ids of those who will be able to decrypt the file
        gpg_ids = cmdline['--gpg-id']
        if not gpg_ids:
            gpg_ids = get_setting('gpg_ids', [])

        # run the generator
        generator = PasswordGenerator(init=cmdline['<name>'], gpg_ids=gpg_ids)


# Search {{{1
class Search(Command):
    NAMES = 'search',
    DESCRIPTION = 'search accounts'
    USAGE = dedent("""
        Search for accounts whose values contain the search text.

        Usage:
            avendesora search <text>
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator()

        # search for accounts that match search criteria
        to_print = []
        for acct in generator.search_accounts(cmdline['<text>']):
            aliases = getattr(acct, 'aliases', [])

            aliases = ' (%s)' % (', '.join(aliases)) if aliases else ''
            to_print += [acct.get_name() + aliases]
        output(cmdline['<text>']+ ':')
        output('    ' + ('\n    '.join(sorted(to_print))))


# Show {{{1
class Show(Command):
    NAMES = 'show', 's',
    DESCRIPTION = 'show an account value'
    USAGE = dedent("""
        Produce an account value. If the value is secret, it is produced only
        temporarily unless --stdout is specified.

        Usage:
            avendesora show [--stdout | --clipboard] [<account> [<field>]]
            avendesora s [--stdout | --clipboard] [<account> [<field>]]

        Options:
            -c, --clipboard         Write output to clipboard rather than stdout.
            -s, --stdout            Write output to the standard output without
                                    any annotation or protections.
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator()

        # determine the account and output specified information
        account_name = cmdline['<account>']
        writer = get_writer(
            bool(account_name), cmdline['--clipboard'], cmdline['--stdout']
        )
        if account_name:
            account = generator.get_account(account_name)
            writer.display_field(account, cmdline['<field>'])
        else:
            # use discovery to determine account
            account_name, script = generator.discover_account()
            account = generator.get_account(account_name)
            writer.run_script(account, script)


# ShowAll {{{1
class Summary(Command):
    NAMES = 'sum', 'summary',
    DESCRIPTION = 'display all account values'
    USAGE = dedent("""
        Show all account values.

        Usage:
            avendesora sum <account>
            avendesora summary <account>
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # run the generator
        generator = PasswordGenerator()

        # determine the account
        account = generator.get_account(cmdline['<account>'])
        account.write_summary()

# Unhide {{{1
class Unhide(Command):
    NAMES = 'unhide', 'reveal',
    DESCRIPTION = 'reveal concealed text'
    USAGE = dedent("""
        Transform concealed text to reveal its original form.

        Usage:
            avendesora [options] unhide <text>
            avendesora [options] reveal <text>

        Options:
            -e <encoding>, --encoding <encoding>
                                    Encoding used when revealing information.
    """).strip()

    @classmethod
    def help(cls):
        encodings = '    ' + '\n    '.join(Conceal.encodings())
        text = dedent("""
            {title}

            {usage}

            Possible encodings include:
            {encodings}

            base64 (default):
                This encoding obscures but does not encrypt the text. It can
                protect text from observers that get a quick glance of the
                encoded text, but if they are able to capture it they can easily
                decode it.
        """).strip()
        return text.format(
            title=title(cls.DESCRIPTION), usage=cls.USAGE, encodings=encodings
        )

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # transform and output the string
        output(Conceal.show(cmdline['<text>'], cmdline['--encoding']))


# Version {{{1
class Version(Command):
    NAMES = 'version',
    DESCRIPTION = 'display Avendesora version'
    USAGE = dedent("""
        Usage:
            avendesora version
    """).strip()

    @classmethod
    def help(cls):
        text = dedent("""
            {title}

            {usage}
        """).strip()
        return text.format(title=title(cls.DESCRIPTION), usage=cls.USAGE)

    @classmethod
    def run(cls, command, args):
        # read command line
        cmdline = docopt(cls.USAGE, argv=[command] + args)

        # output the version
        output('Avendesora version: %s' % __version__)




# Concealers
#
# Classes that are used to conceal the value of a string.

# License {{{1
# Copyright (C) 2016 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from inform import fatal
from binascii import a2b_base64, b2a_base64, Error as BinasciiError
import sys

# Hidden {{{1
class Hidden():
    # This does a simple base64 encoding on the string to hide it from a casual
    # observer. But it is not encrypted. The original value can be trivially
    # recovered from the encoded version.
    def __init__(self, value, secure=True, encoding='utf8'):
        try:
            value = a2b_base64(value)
            self.value = value.decode(encoding)
            self.secure = secure
        except BinasciiError as err:
            import traceback
            exc_type, exc_value, exc_traceback = sys.exc_info()
            filename, lineno = traceback.extract_stack()[-2][:2]
                # context and content are also available, but in this case
                # Hidden is generally instantiated from top-level so the 
                # context is not interesting and the content (the actual line 
                # of code) shown in this case is gibberish (encrypted).
            fatal(
                'invalid value specified to Hidden().',
                culprit=(filename, lineno)
            )

    def _initiate(self, name, account):
        # we don't need to do anything, but having this method marks this value
        # as being confidential
        pass

    def is_secure(self):
        return self.secure

    def __str__(self):
        return self.value

    @staticmethod
    def hide(value, encoding='utf8'):
        value = value.encode(encoding)
        return b2a_base64(value).rstrip().decode('ascii')

    @staticmethod
    def reveal(value, encoding='utf8'):
        value = a2b_base64(value.encode('ascii'))
        return value.decode(encoding)



XK_Arabic_comma = 0x5ac
XK_Arabic_semicolon = 0x5bb
XK_Arabic_question_mark = 0x5bf
XK_Arabic_hamza = 0x5c1
XK_Arabic_maddaonalef = 0x5c2
XK_Arabic_hamzaonalef = 0x5c3
XK_Arabic_hamzaonwaw = 0x5c4
XK_Arabic_hamzaunderalef = 0x5c5
XK_Arabic_hamzaonyeh = 0x5c6
XK_Arabic_alef = 0x5c7
XK_Arabic_beh = 0x5c8
XK_Arabic_tehmarbuta = 0x5c9
XK_Arabic_teh = 0x5ca
XK_Arabic_theh = 0x5cb
XK_Arabic_jeem = 0x5cc
XK_Arabic_hah = 0x5cd
XK_Arabic_khah = 0x5ce
XK_Arabic_dal = 0x5cf
XK_Arabic_thal = 0x5d0
XK_Arabic_ra = 0x5d1
XK_Arabic_zain = 0x5d2
XK_Arabic_seen = 0x5d3
XK_Arabic_sheen = 0x5d4
XK_Arabic_sad = 0x5d5
XK_Arabic_dad = 0x5d6
XK_Arabic_tah = 0x5d7
XK_Arabic_zah = 0x5d8
XK_Arabic_ain = 0x5d9
XK_Arabic_ghain = 0x5da
XK_Arabic_tatweel = 0x5e0
XK_Arabic_feh = 0x5e1
XK_Arabic_qaf = 0x5e2
XK_Arabic_kaf = 0x5e3
XK_Arabic_lam = 0x5e4
XK_Arabic_meem = 0x5e5
XK_Arabic_noon = 0x5e6
XK_Arabic_ha = 0x5e7
XK_Arabic_heh = 0x5e7
XK_Arabic_waw = 0x5e8
XK_Arabic_alefmaksura = 0x5e9
XK_Arabic_yeh = 0x5ea
XK_Arabic_fathatan = 0x5eb
XK_Arabic_dammatan = 0x5ec
XK_Arabic_kasratan = 0x5ed
XK_Arabic_fatha = 0x5ee
XK_Arabic_damma = 0x5ef
XK_Arabic_kasra = 0x5f0
XK_Arabic_shadda = 0x5f1
XK_Arabic_sukun = 0x5f2
XK_Arabic_switch = 0xFF7E



# Abraxas Password Accounts
#
# Responsible for reading and managing the data from the user's accounts file.
#
# Copyright (C) 2013-14 Kenneth S. Kundert and Kale Kundert

# License (fold)
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports (fold)
from __future__ import print_function, division
from abraxas.prefs import (
    DEFAULT_SETTINGS_DIR, DEFAULT_ARCHIVE_FILENAME, DEFAULT_LOG_FILENAME,
    STRING_FIELDS, INTEGER_FIELDS, LIST_FIELDS, LIST_OR_STRING_FIELDS,
    ENUM_FIELDS, SEARCH_FIELDS, PREFER_HTTPS, ACCOUNTS_FILE_INITIAL_CONTENTS,
    XDOTOOL, DEFAULT_AUTOTYPE, TITLE_PATTERNS, URL_PATTERN
)
from fileutils import (
    exists, getExt as get_extension, makePath as make_path,
    getHead as get_head, Execute, ExecuteError
)
import re
import sys
import fnmatch
import traceback

class _Accounts:
    """
    Abraxas Accounts

    Responsible for reading and managing the data from the user's accounts 
    file.
    """

    def __init__(self, path, logger, gpg, template=None, stateless=False):
        self.path = path
        self.logger = logger
        self.gpg = gpg
        self.stateless = stateless
        self.data = None

        if stateless:
            # Use initial accounts so that user has access to basic templates
            imported_data = {}
            exec(ACCOUNTS_FILE_INITIAL_CONTENTS, imported_data)
            self.accounts = imported_data['accounts']
        else:
            # Load the user's accounts file
            self.accounts = self._read_accounts_file()

        if template:
            self.template = self.accounts.get(template, {})
            if not self.template:
                logger.error("%s: template not found." % template)
        else:
            self.template = {}

        self._validate_accounts()
        self._create_aliases()

    def _validate_accounts(self):
        """Validate and repair each account"""
        logger = self.logger

        for ID in self.all_accounts(skip_templates=False):
            if type(ID) != str:
                logger.error('%s: account ID must be a string.' % ID)
            data = self.accounts[ID]

            # check the types of the data in the various fields
            for each in STRING_FIELDS:
                if type(data.get(each, '')) != str:
                    logger.display(
                        ' '.join([
                            "Invalid value for '%s' in %s account (%s)." % (
                                each, ID, data[each]),
                            "Expected string, ignoring."]))
                    del data[each]
            for each in INTEGER_FIELDS:
                if type(data.get(each, 0)) != int:
                    logger.display(
                        ' '.join([
                            "Invalid value for '%s' in %s account (%s)." % (
                                each, ID, data[each]),
                            "Expected integer, ignoring."]))
                    del data[each]
            for each in LIST_FIELDS:
                if type(data.get(each, [])) != list:
                    logger.display(
                        ' '.join([
                            "Invalid value for '%s' in %s account (%s)." % (
                                each, ID, data[each]),
                            "Expected list, ignoring."]))
                    del data[each]
            for each in LIST_OR_STRING_FIELDS:
                if (
                    type(data.get(each, [])) != list and
                    type(data.get(each, '')) != str
                ):
                    logger.display(
                        ' '.join([
                            "Invalid value for '%s' in %s account (%s)." % (
                                each, ID, data[each]),
                            "Expected string or list, ignoring."]))
                    del data[each]
            for key, values in ENUM_FIELDS.items():
                val = data.get(key, '')
                if (val and val not in values):
                    logger.display(
                        ' '.join([
                            "Invalid value for '%s' in %s account (%s)." % (
                                key, ID, data[key]),
                            "Expected one from: %s." % ', '.join(values),
                            "Ignored."]))
                    del data[key]

    def _create_aliases(self):
        """Create dictionary of aliases"""

        def addToAliases(ID, name):
            if name in self.aliases:
                self.logger.display(
                    ' '.join([
                        "Alias %s" % (name),
                        "from account %s" % (
                            ID if name != ID else self.aliases[name]),
                        "duplicates an account name or previous entry,",
                        "ignoring."]))
            else:
                self.aliases[name] = ID

        self.aliases = {}
        for ID in self.all_accounts(skip_templates=False):
            # add ID to the aliases and then add the actual aliases
            data = self.accounts[ID]
            addToAliases(ID, ID)
            for alias in data.get('aliases', []):
                addToAliases(ID, alias)

    def all_accounts(self, skip_templates=True):
        # Get a dictionary of all the fields for each account
        for ID in self.accounts:
            if skip_templates and ID[0] == '=':
                pass
            else:
                yield ID

    def get_fields(self, field):
        # Get a particular field from each account
        for ID, data in self.accounts.items():
            if field in data and data[field]:
                yield (ID, data[field])

    def _read_accounts_file(self):
        if not self.path:
            # There is no accounts file
            self.data = {}
            return self.data
        if not exists(self.path):
            # If file does not exist, look for encrypted versions
            for ext in ['gpg', 'asc']:
                new_path = '.'.join([self.path, ext])
                if exists(new_path):
                    self.path = new_path
                    break

        logger = self.logger
        accounts_data = {}
        try:
            if get_extension(self.path) in ['gpg', 'asc']:
                # Accounts file is GPG encrypted, decrypt it before loading
                with open(self.path, 'rb') as f:
                    decrypted = self.gpg.decrypt_file(f)
                    if not decrypted.ok:
                        logger.error("%s\n%s" % (
                            "%s: unable to decrypt." % (self.path),
                            decrypted.stderr))
                    code = compile(decrypted.data, self.path, 'exec')
                    exec(code, accounts_data)
            else:
                # Accounts file is not encrypted
                with open(self.path) as f:
                    code = compile(f.read(), self.path, 'exec')
                    exec(code, accounts_data)
            if 'accounts' not in accounts_data:
                logger.error(
                    "%s: defective accounts file, 'accounts' not found." %
                        self.path
                )

            # Load additional accounts files
            additional_accounts = accounts_data.get('additional_accounts', [])
            if type(additional_accounts) == str:
                additional_accounts = [additional_accounts]
            for each in additional_accounts:
                more_accounts = {}
                path = make_path(get_head(self.path), each)
                try:
                    if get_extension(path) in ['gpg', 'asc']:
                        # Accounts file is GPG encrypted, decrypt it
                            with open(path, 'rb') as f:
                                decrypted = self.gpg.decrypt_file(f)
                                if not decrypted.ok:
                                    logger.error("%s\n%s" % (
                                        "%s: unable to decrypt." % (path),
                                        decrypted.stderr))
                                    continue
                                code = compile(decrypted.data, path, 'exec')
                                exec(code, more_accounts)
                    else:
                        # Accounts file is not encrypted
                        with open(path) as f:
                            code = compile(f.read(), path, 'exec')
                            exec(code, more_accounts)
                except IOError as err:
                    logger.display('%s: %s.  Ignored' % (
                        err.filename, err.strerror
                    ))
                    continue
                existing_names = set(accounts_data['accounts'].keys())
                new_accounts = more_accounts.get('accounts', {})
                new_names = set(new_accounts.keys())
                names_in_common = sorted(
                    existing_names.intersection(new_names))
                if len(names_in_common) > 2:
                    logger.display(
                        "%s: overrides existing accounts:\n    %s" % (
                            path, ',\n    '.join(sorted(names_in_common))))
                elif names_in_common:
                    logger.display("%s: overrides existing account: %s" % (
                        path, names_in_common[0]))
                accounts_data['accounts'].update(new_accounts)
        except IOError as err:
            logger.error('%s: %s.' % (err.filename, err.strerror))
        except SyntaxError as err:
            traceback.print_exc(0)
            sys.exit()
        self.data = accounts_data
        return accounts_data['accounts']

    def get_log_file(self):
        return self.data.get(
            'log_file',
            make_path(DEFAULT_SETTINGS_DIR, DEFAULT_LOG_FILENAME))

    def get_archive_file(self):
        return self.data.get(
            'archive_file',
            make_path(DEFAULT_SETTINGS_DIR, DEFAULT_ARCHIVE_FILENAME))

    def get_gpg_id(self):
        try:
            return self.data['gpg_id']
        except KeyError:
            if self.data:
                self.logger.error(
                    "'gpg_id' missing from %s (see 'man 5 abraxas')." % (
                        self.path))
            return None

    def all_templates(self):
        # Iterate through  templates
        # Templates are accounts whose ID starts with =.
        for key in self.accounts:
            if key[0] == '=':
                yield key

    class Account:
        """
        Abraxas Account

        Responsible for holding all of the information for a particular 
        account.
        """

        def __init__(self, ID, data):
            self.ID = ID
            self.data = data

        def get_id(self):
            """Return account's ID."""
            return self.ID

        def get_data(self):
            """Return account's data."""
            return self.data

        def get_field(self, field, default=None):
            """
            Return value of a particular account field.

            Arguments:
            field (string)
                Name of desired field.
            default
                Value return if field is not present.
            """
            return self.data.get(field, default)

        def get_master(self, default):
            """Return name of account's master password."""
            return self.data.get('master', default)

        def get_version(self):
            """Return value of account's version field."""
            return self.data.get('version', '')

        def get_security_questions(self):
            """Return list account's security questions."""
            return self.data.get('security questions', [])

        def get_autotype(self):
            return self.data.get('autotype', DEFAULT_AUTOTYPE)

        def get_password_type(self):
            return self.data.get('password-type', 'words')

        def get_num_chars(self, default):
            return self.data.get('num-chars', default)

        def get_num_words(self, default):
            return self.data.get('num-words', default)

        def get_alphabet(self, default):
            return self.data.get('alphabet', default)

        def get_separator(self, default):
            return self.data.get('separator', default)

        def get_prefix(self):
            return self.data.get('prefix', '')

        def get_suffix(self):
            return self.data.get('suffix', '')

    def get_account(self, account_id, level=0):
        if level > 20:
            self.logger.error(
                "%s: too many levels of templates, loop suspected." % (
                    account_id))

        def find_account_id():
            # Uses window title to perform account discovery
            logger = self.logger
            # Account ID was not given by the user.
            # Try to determine it from title of active window.
            # First get the title from the active window.
            try:
                xdotool = Execute([XDOTOOL, 'getactivewindow', 'getwindowname'])
            except ExecuteError as err:
                logger.error(str(err))
            title = xdotool.stdout.strip()
            logger.log('Account Discovery ...')
            logger.log('Focused window title: %s' % title)

            # Look through fields in each account and see if any match.
            #   Title information is separated into components.
            #   Title matches if
            #   - title component matches if given
            #   - host component matches if given
            #   - email component matches if given
            #   - username component matches if given
            #   - account component matches if given
            #   But the following mismatches would invalidate:
            #   - any of the above, if given
            #   - protocol if given
            matches = set([])
            for pattern_name, pattern in TITLE_PATTERNS:
                logger.log('Using title pattern: %s' % pattern_name)
                match = pattern.match(title)
                if match:
                    fields = match.groupdict()
                    logger.log(
                        'Title components:\n    %s' % '\n    '.join([
                            '%s: %s' % (key, val)
                            for key, val in fields.items()
                        ])
                    )
                    required_protocol = None
                    for ID, account in self.accounts.items():
                        logger.debug('Trying account: %s' % ID)
                        windows = account.get('window', [])
                        if type(windows) == str:
                            windows = [windows]

                        match_found = False
                        reasons = []
                        for key in sorted(
                            fields.keys(), key=lambda x: x == 'protocol'
                        ):
                            # The above has a special sort that assures 
                            # protocol is processed last.
                            value = fields[key]
                            if not value:
                                continue
                            elif key == 'title':
                                for each in windows:
                                    if fnmatch.fnmatch(value, each):
                                        match_found = True
                                        logger.debug('    title matches')
                                        reasons += ['title matches']
                                        break
                                else:
                                    if windows:
                                        logger.debug('    title mismatch')
                                        break
                            elif key == 'host':
                                urls = account.get('url', [])
                                if type(urls) == str:
                                    urls = [urls]
                                for url in urls:
                                    logger.debug('    account url: %s' % url)
                                    match = URL_PATTERN.match(url)
                                    if match:
                                        url = match.groupdict()
                                        logger.debug(
                                            '    url components:\n        %s' %
                                                '\n        '.join([
                                                    '%s: %s' % (k, v)
                                                    for k, v in url.items()
                                                ])
                                        )
                                    else:
                                        url = {}
                                    if fnmatch.fnmatch(
                                        value, url.get('host', '')
                                    ):
                                        match_found = True
                                        logger.debug('    host matches')
                                        reasons += ['host matches']
                                        required_protocol = url['protocol']
                                        if required_protocol:
                                            required_protocol = (
                                                required_protocol.lower())
                                        break
                                else:
                                    logger.debug('    host mismatch')
                                    break
                            elif key == 'username':
                                if key == account.get('username'):
                                    match_found = True
                                    logger.debug('    username matches')
                                    reasons += ['username matches']
                                else:
                                    logger.debug('    username mismatch')
                                    break
                            elif key == 'email':
                                if key == account.get('email'):
                                    logger.debug('    email matches')
                                    reasons += ['email matches']
                                    match_found = True
                                else:
                                    logger.debug('    email mismatch')
                                    break
                            elif key == 'account':
                                if key == account.get('account'):
                                    logger.debug('    account matches')
                                    reasons += ['account matches']
                                    match_found = True
                                else:
                                    logger.debug('    account mismatch')
                                    break
                            elif key == 'protocol':
                                if PREFER_HTTPS and not required_protocol:
                                    required_protocol = 'https'
                                if (
                                    required_protocol and
                                    value.lower() != required_protocol
                                ):
                                    logger.debug('    protocol mismatch')
                                    if match_found:
                                        if required_protocol == 'https':
                                            # this is the last test, and if 
                                            # a match is found but rejected 
                                            # because we are expecting https, 
                                            # warn the user that the page is 
                                            # not encrypted
                                            import abraxas.dialog
                                            abraxas.dialog.show_error_dialog(
                                                ' '.join([
                                                    "Account '%s' expects" % ID,
                                                    "page to be encrypted."
                                                ])
                                            )
                                    break
                        else:
                            if match_found:
                                logger.debug('    match!')
                                successful_reasons = reasons
                                matches.add(ID)
                            else:
                                logger.debug('    no fields to match')
                if matches:
                    # don't go through again if a match has already been found
                    break

            # Only a single match is allowed.
            logger.log(
                "Window title matches the following accounts: '%s'." % (
                    "' '".join(matches)))
            if len(matches) == 1:
                match = matches.pop()
                self.logger.log(
                    "'%s' account selected because %s." % (
                        match, ', '.join(successful_reasons)))
                return match
            elif matches:
                from abraxas.dialog import show_list_dialog
                account = show_list_dialog(sorted(matches))
                if account:
                    logger.log("User selected '%s' account." % account)
                    return account
            logger.error("Cannot determine desired account ID.%s" % (
                ("\nExamine '%s' for the details." % logger.logfile)
                if logger.logfile
                else ""
            ))

        # Validate account_id
        if not account_id and not self.stateless:
            # User did not specify account ID on the command line.
            account_id = find_account_id()
        try:
            account_id = self.aliases[account_id]
            account = self.accounts[account_id]
        except KeyError:
            account = self.template
            if not self.stateless:
                self.logger.display(
                    "Warning: account '%s' not found." % account_id)

        # Get information from template
        template = account.get('template', None)
        if template:
            data = self.get_account(template, level=level+1).get_data()
        else:
            data = {}

        # Override template information with that from the account
        data.update(account)

        return _Accounts.Account(account_id, data)

    @staticmethod
    def _inID(pattern, ID):
        return bool(pattern.search(ID))

    @staticmethod
    def _inAliases(pattern, acct):
        for each in acct.get('aliases', ''):
            if pattern.search(each):
                return True
        return False

    def _inSearchField(self, pattern, ID, acct):
        for each in SEARCH_FIELDS:
            try:
                value = acct.get(each, '')
                if type(value) is list:
                    for every in value:
                        if pattern.search(every):
                            return True
                else:
                    if pattern.search(value):
                        return True
            except TypeError:
                self.logger.display(
                    "%s %s: field is of wrong type" % (ID, each))
        return False

    def find_accounts(self, target):
        """Iterate through accounts that match target.

        Look for target in account ID and aliases only.
        """

        pattern = re.compile(target, re.I)
        for ID in self.all_accounts():
            if (
                    self._inID(pattern, ID) or
                    self._inAliases(pattern, self.accounts[ID])):
                yield ID, self.accounts[ID].get('aliases', [])

    # Search accounts
    def search_accounts(self, target):
        """Iterate through accounts that match target.

        Look for target in account ID, aliases, and various fields.
        """
        pattern = re.compile(target, re.I)
        for ID in self.all_accounts():
            acct = self.accounts[ID]
            if (
                    self._inID(pattern, ID) or
                    self._inAliases(pattern, acct) or
                    self._inSearchField(pattern, ID, acct)):
                yield ID, self.accounts[ID].get('aliases', [])

# vim: set sw=4 sts=4 et:



XK_Aogonek = 0x1a1
XK_breve = 0x1a2
XK_Lstroke = 0x1a3
XK_Lcaron = 0x1a5
XK_Sacute = 0x1a6
XK_Scaron = 0x1a9
XK_Scedilla = 0x1aa
XK_Tcaron = 0x1ab
XK_Zacute = 0x1ac
XK_Zcaron = 0x1ae
XK_Zabovedot = 0x1af
XK_aogonek = 0x1b1
XK_ogonek = 0x1b2
XK_lstroke = 0x1b3
XK_lcaron = 0x1b5
XK_sacute = 0x1b6
XK_caron = 0x1b7
XK_scaron = 0x1b9
XK_scedilla = 0x1ba
XK_tcaron = 0x1bb
XK_zacute = 0x1bc
XK_doubleacute = 0x1bd
XK_zcaron = 0x1be
XK_zabovedot = 0x1bf
XK_Racute = 0x1c0
XK_Abreve = 0x1c3
XK_Lacute = 0x1c5
XK_Cacute = 0x1c6
XK_Ccaron = 0x1c8
XK_Eogonek = 0x1ca
XK_Ecaron = 0x1cc
XK_Dcaron = 0x1cf
XK_Dstroke = 0x1d0
XK_Nacute = 0x1d1
XK_Ncaron = 0x1d2
XK_Odoubleacute = 0x1d5
XK_Rcaron = 0x1d8
XK_Uring = 0x1d9
XK_Udoubleacute = 0x1db
XK_Tcedilla = 0x1de
XK_racute = 0x1e0
XK_abreve = 0x1e3
XK_lacute = 0x1e5
XK_cacute = 0x1e6
XK_ccaron = 0x1e8
XK_eogonek = 0x1ea
XK_ecaron = 0x1ec
XK_dcaron = 0x1ef
XK_dstroke = 0x1f0
XK_nacute = 0x1f1
XK_ncaron = 0x1f2
XK_odoubleacute = 0x1f5
XK_udoubleacute = 0x1fb
XK_rcaron = 0x1f8
XK_uring = 0x1f9
XK_tcedilla = 0x1fe
XK_abovedot = 0x1ff



# Xlib.xobject.resource -- any X resource object
#
#    Copyright (C) 2000 Peter Liljenberg <petli@ctrl-c.liu.se>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from Xlib.protocol import request

class Resource:
    def __init__(self, display, rid, owner = 0):
        self.display = display
        self.id = rid
        self.owner = owner

    def __resource__(self):
        return self.id

    def __eq__(self, obj):
        if isinstance(obj, Resource):
            if self.display == obj.display:
                return self.id == obj.id
            else:
                return self.display == obj.display
        else:
            return id(self) == id(obj)

    def __hash__(self):
        return int(self.id)

    def __str__(self):
        return '%s(0x%08x)' % (self.__class__, self.id)

    def __repr__(self):
        return '<%s 0x%08x>' % (self.__class__, self.id)

    def kill_client(self, onerror = None):
        request.KillClient(display = self.display,
                           onerror = onerror,
                           resource = self.id)
